#! /usr/bin/env python3
#
# Client Example for webrepl and dummy
#
# This file is part of Zepl Broker: https://gitlab.com/zepl1/zepl-broker
# (C)2020 Leonard Pollak <leonardp@tr-host.de>
#
# SPDX-License-Identifier:    AGPL-3
import sys, json, zmq, asyncio

SUBSCRIPTIONS = [''] # http://wiki.zeromq.org/whitepapers:message-matching

def main():
    """Command line tool, entry point"""

    import argparse

    parser = argparse.ArgumentParser(
        description='Zepl Client')

    subparsers = parser.add_subparsers(help='commands')
    subparser_log = subparsers.add_parser('log', help='log listener')
    subparser_dev = subparsers.add_parser('dev', help='send to device/runner')
    subparser_ctrl = subparsers.add_parser('ctrl', help='broker ctrl interface')

    subparser_log.add_argument(
        'log',
        help='log commands',
        nargs='?',
        choices=('all', ),
        default='all')

    subparser_dev.add_argument(
        'dev',
        help='dev commands',
        nargs='?',
        choices=('greet', 'sync', 'run'),
        default='greet')

    subparser_dev.add_argument(
        'dev_id',
        help='dev_id of target')

    subparser_dev.add_argument(
        'app',
        nargs='?',
        help='input for mp example app -- single character')

    subparser_ctrl.add_argument(
        'ctrl',
        help='ctrl commands',
        nargs='?',
        choices=('get', 'add'),
        default='get')

    parser.add_argument(
        '--broker',
        nargs='?',
        help='hostname or ip of broker',
        default='localhost')

    parser.add_argument(
        '--config',
        nargs='?',
        help='device config',
        default=False)

    parser.add_argument(
        '-d',
        help='dummy: spawns dummy devices from config',
        action='store_true',
        default=False)

    args = parser.parse_args()

    if args.config:
        with open(args.config, "r") as f:
            cfg = json.load(f)

    ctx = zmq.Context()
    if args.d:
        from zepl_broker import DummyClient
        client = DummyClient(ctx, args.broker)
    else:
        from zepl_broker import WebReplClient
        client = WebReplClient(ctx, args.broker)

    print(f'-- Starting Zepl Client --')
    try:
        if 'log' in args:
            for dev_id in SUBSCRIPTIONS:
                client.log_sub(dev_id)
            while True:
                client.print_log()
        elif 'dev' in args:
            if 'greet' in args.dev:
                client.greet(args.dev_id)
            else:
                if 'sync' in args.dev:
                    client.sync_files(args.dev_id)
                if 'run' in args.dev:
                    client.run_app(args.dev_id, args.app)
        elif 'ctrl' in args:
            if 'get' in args.ctrl:
                client.show_devices()
            elif 'add' in args.ctrl:
                if args.d:
                    ## spawn dummy devices locally
                    from zepl_device.devices.dummy import DummyDevice
                    dctx = zmq.asyncio.Context()
                    for dev_id in cfg.keys():
                        d = DummyDevice(dctx, cfg[dev_id])
                        d.start()
                        client.add_device(dev_id, cfg[dev_id])
                    loop = asyncio.get_event_loop()
                    loop.run_forever()
                else:
                    for dev_id in cfg.keys():
                        client.add_device(dev_id, cfg[dev_id])
    except Exception as e:
        print(e)

# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
if __name__ == '__main__':
    main()



# spawn dummy devices and start them in a seperate process
import zmq
from multiprocessing import Process
from zepl_broker import DummyClient
from time import sleep

broker_ip = '192.168.0.5' # raspberrypi
dummy_machine_ip = '192.168.0.10' # desktop
dummy_count = 30
task_num = 2

# dummy devices
dummy_cfg = {}
port = 5000
d_template = {
    'dev_ip': dummy_machine_ip,
    'dev_port': port,
    'proto': 'dummy-example',
    'dev_pw': 'asdf'
}
for d in range(dummy_count):
    port += 1
    d_template['dev_port'] = port
    dummy_cfg[f'dummy{d}'] = {}
    dummy_cfg[f'dummy{d}'].update(d_template)


def dd_proc(broker, cfg):
    import zmq.asyncio, asyncio
    from zepl_device.devices.dummy import DummyDevice
    print(cfg)
    ctx = zmq.asyncio.Context()
    for dev_id in cfg.keys():
        d = DummyDevice(ctx, cfg[dev_id])
        d.start()
    loop = asyncio.get_event_loop()
    loop.run_forever()

dev_proc = Process(target=dd_proc, args=(broker_ip, dummy_cfg))
dev_proc.start()

## client stuff
ctx = zmq.Context()
client = DummyClient(ctx, broker_ip)
for dev_id in dummy_cfg.keys():
    client.add_device(dev_id, dummy_cfg[dev_id])

# start sender tasks
for dev_id in dummy_cfg.keys():
    for x in range(task_num):
        sleep(0.05)
        client.start_task(dev_id, 'device', f'task{x}', 'a', arg={'device': f'task{x}'})
        sleep(0.05)
        client.start_task(dev_id, 'device', f'task{x}', 'b', arg={'device': f'task{x}'})
        sleep(0.05)
        client.start_task(dev_id, 'runner', f'task{x}', 'a', arg={'runner': f'task{x}'})
        sleep(0.05)
        client.start_task(dev_id, 'runner', f'task{x}', 'b', arg={'runner': f'task{x}'})

# msg counter
print('>>> Starting message counter...')
print(f'Number of sender tasks is: {dummy_count*task_num*4} (1/2 devices and runners each.')
from time import perf_counter
client.log_sub('')
avg = 3
s_counter, msg_counter, avg_counter = [0 ,0 ,0]
tic = perf_counter()
while True:
    client.cnt_log()
    msg_counter += 1
    toc = perf_counter()
    if (toc-tic) > 1:
        tic = perf_counter()
        s_counter += 1
        avg_counter += msg_counter
        if s_counter > avg:
            print(f'({avg}s): {int(avg_counter/avg)} msg/s',  end='\r', flush=True)
            avg_counter = 0
            s_counter = 0
        msg_counter = 0



from machine import Pin, TouchPad, I2C, SDCard
#from machine import Pin, TouchPad, I2C, I2S, SDCard
import sys, os, machine
import esp32, neopixel
from time import sleep

NP_PIN = Pin(0)
led = neopixel.NeoPixel(NP_PIN, 1)

try:
    os.mount(SDCard(), "/sd")
    led.fill((0,255,0))
    print('LOG SD Card mounted.')
except:
    led.fill((255,0,0))
    print('ERROR Could not Mount SD Card!')

led.write()

print('LOG app_b finished')



import json, sys, os, machine, binascii, gc, hashlib, select
import neopixel as np
import time

# initialization
p = select.poll()
p.register(sys.stdin, select.POLLIN)

dev_id = binascii.hexlify(machine.unique_id()).decode()
NP_PIN = machine.Pin(0)

file_hashtab = {}
for f in os.listdir():
    with open(f, 'r') as fd:
        hasher = hashlib.sha1()
        lines = fd.readlines()
        hasher.update("".join(lines))
        file_hashtab[f] = binascii.hexlify(hasher.digest())
    del hasher

hashtab_json = json.dumps(file_hashtab)
del file_hashtab
gc.collect()

print('Waiting for input...')
led = np.NeoPixel(NP_PIN, 1)
led.fill((0,255,0))
led.write()

act=[]
ctrl_input=None
while True:
    while p.poll(100):
        ctrl_input = sys.stdin.read(1)
    #print('CTRL_INPUT {}'.format(ctrl_input))

    if ctrl_input == '1':
        print('DEV_MAC {}'.format(dev_id))
    elif ctrl_input == '2':
        print('F_HTAB {}'.format(hashtab_json))
    elif ctrl_input == '3':
        print('HEART BEAT')
        print('LOG sleeping 5 seconds')
        time.sleep(5)
    elif ctrl_input == 'a':
        try:
            import app_a
        except:
            print('ERROR problem in a')
    elif ctrl_input == 'b':
        try:
            import app_b
        except:
            print('ERROR problem in b')
    else:
        if ctrl_input:
            print('ERROR unknown control input: {}'.format(ctrl_input))



#! /usr/bin/env python3
#
# example client implementations
#
# This file is part of Zepl Broker: https://gitlab.com/zepl1/zepl-broker
# (C)2020 Leonard Pollak <leonardp@tr-host.de>
#
# SPDX-License-Identifier:    AGPL-3
import json, zmq
from pprint import pprint

class ClientLog:
    def cnt_log(self):
        self.log.recv_multipart()

    def print_log(self):
        dev_id, prefix, msg = self.log.recv_multipart()
        print(f'{dev_id} -- {prefix} --  {msg}')

    def log_sub(self, dev_id):
        self.log.subscribe(dev_id.encode())

    def log_unsub(self, dev_id):
        self.log.unsubscribe(dev_id.encode())


class ClientCtrl:
    def show_devices(self):
        print_str = 'Remote Devices: '
        cmd = ['get']
        self.req_rep(self.ctrl, print_str, cmd)

    def add_device(self, dev_id, cfg):
        print_str = f'Added {dev_id}: '
        cmd = ['add', dev_id, json.dumps(cfg)]
        self.req_rep(self.ctrl, print_str, cmd)

    def rem_device(self, dev_id):
        print_str = f'Removed {dev_id}: '
        cmd = ['rem', dev_id]
        self.req_rep(self.ctrl, print_str, cmd)


class Client(ClientLog, ClientCtrl):
    def __init__(self, ctx, broker_ip):
        super().__init__()

        self.log = ctx.socket(zmq.SUB)
        self.log.connect(f'tcp://{broker_ip}:9000')

        self.dev = ctx.socket(zmq.DEALER)
        self.dev.connect(f'tcp://{broker_ip}:9001')

        self.ctrl = ctx.socket(zmq.DEALER)
        self.ctrl.connect(f'tcp://{broker_ip}:9002')

    def req_rep(self, sock, print_str, cmd):
        cmd = [s.encode() for s in cmd]
        sock.send_multipart(cmd)
        try:
            resp = sock.recv_multipart()
            print(f'{print_str}: ', resp)
        except Exception as e:
            print('req_rep exception: ', e)


class DummyDevice:
    def greet(self, dev_id):
        print_str = f'greeting {dev_id} runner -- '
        cmd = [dev_id, 'runner', 'greetings']
        self.req_rep(self.dev, print_str, cmd)

        print_str = f'greeting {dev_id} device -- '
        cmd = [dev_id, 'device', 'greetings']
        self.req_rep(self.dev, print_str, cmd)

    def start_task(self, dev_id, target, task_name, task_id, arg={}):
        print_str = f'starting {dev_id}:{target}:{task_name} -- '
        cmd = [dev_id, target, 'start', task_name, task_id, json.dumps(arg)]
        self.req_rep(self.dev, print_str, cmd)

    def stop_task(self, dev_id, target, task_name):
        print_str = f'stopping {dev_id}:{target}:{task_name} -- '
        cmd = [dev_id, target, 'stop', task_name]
        self.req_rep(self.dev, print_str, cmd)


class WebReplDevice:
    def greet(self, dev_id):
        print_str = f'greeting {dev_id} -- '
        cmd = [dev_id, 'runner', 'greetings']
        self.req_rep(self.dev, print_str, cmd)

    def run_app(self, dev_id, app_char):
        print_str = f'running app {app_char} on {dev_id} -- '
        cmd = [dev_id, 'runner', 'run', app_char]
        self.req_rep(self.dev, print_str, cmd)

    def sync_files(self, dev_id):
        print_str = f'syncing files {dev_id} -- '
        cmd = [dev_id, 'runner', 'sync']
        self.req_rep(self.dev, print_str, cmd)


class WebReplClient(Client, WebReplDevice):
    def __init__(self, ctx, broker_ip):
        super().__init__(ctx, broker_ip)


class DummyClient(Client, DummyDevice):
    def __init__(self, ctx, broker_ip):
        super().__init__(ctx, broker_ip)



def scan_good_wifi(sta_if, wifi_configs):
    import ubinascii

    wifi_scan = sta_if.scan()
    wifi_scan =  [[w[0].decode(), ubinascii.hexlify(w[1]).decode(),w[2],w[3],w[4],w[5]] for w in wifi_scan]
    wifi_scan_sorted = sorted(wifi_scan, key=lambda e:e[3], reverse=True)
    good_wifis = [[w[0],wifi_configs[w[0]]] for w in wifi_scan_sorted if (w[0] in list(wifi_configs.keys()))]
    return good_wifis

def do_connect(timeout = 10):
    import json, network, time
    try:
        with open('wifi.secret','r') as f:
            wifi_configs = json.load(f)
    except:
        return

    sta_if = network.WLAN(network.STA_IF)
    sta_if.active(True)
    if not sta_if.isconnected():
        print('connecting to network...')
        good_wifis = scan_good_wifi(sta_if, wifi_configs)
        for good_wifi in good_wifis:
            sta_if.connect(good_wifi[0], good_wifi[1])
            startT = time.time()
            while not sta_if.isconnected():
                time.sleep(0.1)
                if (time.time()-startT)>timeout:
                    break
            if sta_if.isconnected():
                break

    print('network config:', sta_if.ifconfig())
    return sta_if

def disable_wifi(wifi):
    if wifi:
        wifi.disconnect()
        wifi.active(False)
    else:
        print('why u do dis?')



#
# This file is part of Zepl Broker: https://gitlab.com/zepl1/zepl-broker
# (C)2020 Leonard Pollak <leonardp@tr-host.de>
#
# SPDX-License-Identifier:    AGPL-3
import json
import asyncio
import zmq.asyncio
from pprint import pprint

from zepl_device import ZeplDevice

class ZeplBroker():
    uri_dev_in = 'inproc://dev_in'
    uri_dev_out = 'inproc://dev_out'

    ctx = zmq.asyncio.Context(io_threads=4)

    dev_out = ctx.socket(zmq.SUB)
    dev_out.bind(uri_dev_out)
    dev_out.subscribe(b'')

    dev_in = ctx.socket(zmq.PUB)
    dev_in.bind(uri_dev_in)

    log_out = ctx.socket(zmq.PUB)
    device = ctx.socket(zmq.ROUTER)
    ctrl = ctx.socket(zmq.ROUTER)

    _alive = False

    def __init__(self, cfg=None, uri_log_out='tcp://0.0.0.0:9000', uri_device='tcp://0.0.0.0:9001', uri_ctrl='tcp://0.0.0.0:9002'):
        self.loop = asyncio.get_event_loop()
        self.log_out.bind(uri_log_out)
        self.device.bind(uri_device)
        self.ctrl.bind(uri_ctrl)
        self.dev_map = {}

    def start(self):
        self._alive = True
        self.ctrl_task = asyncio.ensure_future(self.broker_switch())

    def add_device(self, dev_id, cfg):
        try:
            device = ZeplDevice(self.ctx, dev_id, self.uri_dev_in, self.uri_dev_out, cfg)
        except Exception as e:
            print(e)
        device.start()
        self.dev_map[dev_id] = device

    def remove_device(self, dev):
        """collect garbage?
        """
        self.dev_map[dev].stop()
        del self.dev_map[dev]

    async def broker_switch(self):
        poller = zmq.asyncio.Poller()
        poller.register(self.dev_out, zmq.POLLIN)
        poller.register(self.device, zmq.POLLIN)
        poller.register(self.ctrl, zmq.POLLIN)

        self._alive = True
        while self._alive:
            socks = dict(await poller.poll(100))

            if self.device in socks.keys():
                try:
                    raw = await self.device.recv_multipart()
                    raw[0], raw[1] = raw[1], raw[0]
                    await self.dev_in.send_multipart(raw)
                except Exception as e:
                    print(f'broker(device) dropping msg: {raw} -- {e}')

            if self.dev_out in socks.keys():
                try:
                    topic, msg_type, prefix, msg = await self.dev_out.recv_multipart()
                except Exception as e:
                    print(f'broker(dev_out) dropping msg: {raw} -- {e}')

                if msg_type == b'log':
                    await self.log_out.send_multipart([topic, prefix, msg])
                elif msg_type == b'ctrl':
                    await self.device.send_multipart([prefix, msg])
                else:
                    #print(f'broker(dev_out) dropping msg: {topic}{msg_type}{raw}')
                    pass

            if self.ctrl in socks.keys():
                try:
                    raw = await self.ctrl.recv_multipart()
                    resp = self.ctrl_handler(raw)
                    await self.ctrl.send_multipart(resp)
                except Exception as e:
                    print(f'broker(ctrl) dropping msg: {raw} -- {e}')

    def ctrl_handler(self, raw):
        route = raw.pop(0)
        cmd = raw.pop(0)
        if cmd == b'get':
            resp = [d.encode() for d in self.dev_map]
            if len(resp) == 0:
                resp = b'no devices active'
        elif cmd == b'add':
            try:
                dev_id = raw.pop(0)
                cfg = json.loads(raw.pop(0).decode())
                self.add_device(dev_id.decode(), cfg)
                resp = b'success'
            except:
                resp = b'error adding device'
        elif cmd == b'rem':
            try:
                d = raw.pop(0)
                self.remove_device(d)
                resp = b'success'
            except:
                resp = b'error removing device'
        else:
            resp = b'unknown cmd'
        if isinstance(resp, bytes):
            resp = [route, resp]
        else:
            resp = [route, *resp]
        return resp

    def stop(self):
        self._alive = False
        devs = [d for d in self.dev_map.keys()]
        for d in devs:
            self.remove_device(d)
        self.ctrl_task.cancel()



from setuptools import setup, find_namespace_packages

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="zepl-broker",
    version="0.9.4",
    license = 'AGPL-3.0-or-later',
    author="Leonard Pollak",
    author_email="leonardp@tr-host.de",
    description="Zepl Broker",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://gitlab.com/zepl1/zepl-broker",
    python_requires='>=3.7',
    install_requires=['pyzmq', 'zepl-device'],
    classifiers=[
      "Programming Language :: Python :: 3",
      "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
    ],
    package_dir = {'': 'src'},
    packages=find_namespace_packages(where='src'),
    entry_points = {
        'console_scripts': [
            'zepl-broker = zepl_broker.broker_cli:main',
            'zepl-client = zepl_broker.client_cli:main'
        ]
    }
)



import sys, os, machine
import select
import esp32
import neopixel

from time import sleep

NP_PIN = machine.Pin(0)
led = neopixel.NeoPixel(NP_PIN, 1)

led.fill((25,25,25))
led.write()

print('LOG app_a finished')
sleep(2)



#!/usr/bin/env python
#
# This file is part of Zepl Broker: https://gitlab.com/zepl1/zepl-broker
# (C)2020 Leonard Pollak <leonardp@tr-host.de>
#
# SPDX-License-Identifier:    AGPL-3

import sys, os, hashlib, binascii
import datetime as dt
import json

import asyncio
import zmq.asyncio

from pprint import pprint

from zepl_broker import ZeplBroker

def main():
    print(f'-- Starting Zepl Broker --')

    lc = ZeplBroker()
    try:
        loop = asyncio.get_event_loop()
        lc.start()
        loop.run_forever()
    except KeyboardInterrupt:
        pass
    finally:
        lc.stop() # how to clean shutdown?????
        print(f'-- Zepl Broker Terminated --')

if __name__ == '__main__':
    main()
